# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from datetime import datetime, timedelta

import click
from caldav.lib import error

from caldavctl import dav, get_name
from caldavctl.click_utils import create_compatibility_check
from caldavctl.event_builder import event_builder, parse_event
from caldavctl.object_parser import ParserError
from caldavctl.templates import render
from caldavctl.utils import calmonth, edit_text_with_editor, to_datetime

# Configuration

EMPTY_EVENT = '''# Mandatory
DTSTART: %(start)s
DTEND: %(end)s

%(cal)s

# Recommended
SUMMARY:

# Optional
LOCATION:
CATEGORIES:
TIMEZONE:
PRIORITY: 5
ALARM: P0D
DESCRIPTION: [[ ]]

# NOTES:
#
#   * Date and time:
#       * The dates must be in iso format, for instance: 2024-12-29 13:45;
#       * The timezone used is the one defined by default or the one defined in
#         the "TIMEZONE" key;
#       * You can use the command "caldavctl utils list-timezones" to list the
#         available time zones;
#   * Categories: the categories are a comma separated list;
#   * Priority: 1 - highest priority, 9 - lowest priority
#   * Description: The description can be multi line, just make sure it's
#     delimited by [[ ]].
#   * Alarms:
#       * By default an alarm is defined the moment the event begins;
#       * The alarms are defined using:
#           ALARM: <trigger> [<description>]
#       * The trigger is defined according to:
#           https://datatracker.ietf.org/doc/html/rfc5545#autoid-38
#       * For example, to trigger an alarm 30 minutes before the event start, use:
#           ALARM: -PT30M
#       * The alarm description is optional
'''


# Commands:

exclusive_option = create_compatibility_check(
    ['today', 'week', 'dtstart', 'dtend', 'day'],  # Exclusive options
    [['dtstart', 'dtend']])  # Except these two that can be used together


@click.command('list', options_metavar='[options]')
@click.option('-t', '--today',
              is_flag=True, show_default=True, default=False,
              callback=exclusive_option, help='Today\'s events',)
@click.option('-w', '--week',
              default=None, metavar='<offset>', type=int,
              callback=exclusive_option, help='Week events')
@click.option('-s', '--dtstart',
              callback=exclusive_option, help='Date range, start date', metavar='<yyyy-mm-dd>')
@click.option('-e', '--dtend',
              callback=exclusive_option, help='Date range, end date', metavar='<yyyy-mm-dd>')
@click.option('-d', '--day',
              callback=exclusive_option, help='Events on the day', metavar='<yyyy-mm-dd>')
@click.option('-sd', '--show-description',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s description.')
@click.option('-si', '--show-uid',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s UID.')
@click.option('-tf', '--template-file', default='event.txt',
              help='Template used to format the output', metavar='<file>')
@click.pass_obj
def list_events(context, today, week, dtstart, dtend, day, show_description, show_uid, template_file):
    '''
    List events from the server list

    The date range can be specified in several ways. The options are self
    explanatory except the --week option. This option takes an integer
    argument, if we use '0' the time range will be the current week. If we use
    '1' the time range will correspond to next week, likewise, '-1' corresponds
    to last week or '-2' to two weeks ago.
    '''
    tz = context['config'].tz()
    start_date = None
    end_date = None

    if today:
        start_date = datetime.now().replace(hour=0, minute=0, second=0, tzinfo=tz)
        end_date = datetime.now().replace(hour=23, minute=59, second=59, tzinfo=tz)
    elif week is not None:
        today_date = datetime.now().replace(hour=0, minute=0, second=0, tzinfo=tz)
        start_date = today_date - timedelta(days=today_date.weekday())
        end_date = start_date + timedelta(days=6)
        end_date = end_date.replace(hour=23, minute=59, second=59)
        # Apply week offset
        start_date += timedelta(weeks=week)
        end_date += timedelta(weeks=week)
    elif dtstart and dtend:
        if dtstart:
            start_date = to_datetime(dtstart, tz)
        else:
            start_date = datetime(1900, 1, 1, tzinfo=tz)
        if dtend:
            end_date = to_datetime(dtend, tz)
        else:
            end_date = datetime(5000, 1, 1, tzinfo=tz)
    elif day:
        try:
            dt = datetime.fromisoformat(day)
            start_date = dt.replace(hour=0, minute=0, second=0, tzinfo=tz)
            end_date = dt.replace(hour=23, minute=59, second=59, tzinfo=tz)
        except ValueError:
            raise click.UsageError(f'Invalid date "{day}"')
    else:
        raise click.UsageError('You have to specify a time frame, for example with -t or -w.')

    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        # Check if the calendar supports events:
        if 'VEVENT' not in calendar.get_supported_components():
            raise click.UsageError('This calendar does not support events.')

        # Get events in time range
        events = calendar.search(
            start=start_date,
            end=end_date,
            event=True,
            expand=True,
            sort_keys=['dtstart']
        )
        event_list = []
        for event in events:
            ev = event.icalendar_component
            date = ev.get('dtstart').dt
            event_context = {
                'date': date.strftime('%Y-%m-%d %H:%M') if isinstance(date, datetime) else f'{date} (day)',
                'summary': ev.get('summary', ''),
                'description': ev.get('description', ''),
                'uid': ev.get('uid')
            }
            event_list.append(event_context)

        template_context = {
            'show_uid': show_uid,
            'show_description': show_description,
            'event_list': event_list,
            'newline': '\n'
        }

        txt = render(get_name(), template_file, template_context)
        click.echo(txt.strip())


@click.command('create', options_metavar='[options]')
@click.option('-f', '--file', type=click.File('r'), default=sys.stdin, help='Create event from this file', metavar='<file>')
@click.option('-e', '--edit', is_flag=True, default=False, help='Edit the event on a text editor?')
@click.option('--dry-run', is_flag=True, default=False, help='Show only the iCalendar generated')
@click.pass_obj
def create_event(context, file, edit, dry_run):
    '''
    Create new event

    The event data can be read from stdin or from a file. Optionally the event
    can be edited using the default $EDITOR. If the option to edit is enabled
    and no file is defined using -f/--file, then an empty event is opened.
    '''
    tz = context['config'].tz()
    # Read the event definition
    if edit and file is sys.stdin:
        tomorrow = datetime.now() + timedelta(days=1)
        event_string = EMPTY_EVENT % {
            'tz': f'{tz}',
            'start': tomorrow.replace(hour=9, minute=0, second=0).strftime('%Y-%m-%d %H:%M'),
            'end': tomorrow.replace(hour=9, minute=45, second=0).strftime('%Y-%m-%d %H:%M'),
            'cal': '\n'.join([('# ' + ln).strip() for ln in calmonth(tomorrow.year, tomorrow.month, 3)])
        }
    else:
        event_string = file.read()

    if edit:
        # Edit the file in $VISUAL or $EDITOR
        old_string = event_string
        event_string = edit_text_with_editor(event_string, suffix='.calendar_event')
        if old_string == event_string:
            click.confirm('No changes made. Do you want to continue?', abort=True)

    try:
        event_data = parse_event(event_string, tz)
    except ParserError as msg:
        raise click.UsageError(msg)

    event_ics = event_builder(event_data, tz)

    if dry_run:
        click.echo(event_ics)
    else:
        # Save the event to the calendar
        _, server = context['config'].get_server()
        calendar_id = context['config'].get_calendar()
        with dav.caldav_calendar(server, calendar_id) as calendar:
            # Check if the calendar supports events:
            if 'VEVENT' not in calendar.get_supported_components():
                raise click.UsageError('This calendar does not support events.')
            try:
                calendar.save_event(event_ics)
            except error.AuthorizationError as msg:
                raise click.UsageError(f'Error saving the event to the server: {msg}')

        click.echo('Event created successfully')


@click.command('delete', options_metavar='[options]')
@click.argument('event_id', metavar='<uid>')
@click.pass_obj
def delete_event(context, event_id):
    '''Delete an event on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_event(server, calendar_id, event_id) as event:
        event.delete()
    print('Event deleted')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
CalDAV tool
'''

import click

import caldavctl.calendar_dav as calendar
import caldavctl.event as event
import caldavctl.todo as todo
import caldavctl.journal as journal
import caldavctl.ics as ics
import caldavctl.backup as backup
import caldavctl.utilities as utils

from caldavctl import get_version
from caldavctl.config import Config


def print_version(context, param, value):
    # See https://click.palletsprojects.com/en/stable/options/#callbacks-and-eager-options
    if not value or context.resilient_parsing:
        return
    click.echo(f'caldavctl v{get_version()}')
    context.exit(0)


# General options

@click.group(options_metavar='[options]')
@click.option('-c', '--config', 'config_file', envvar='CALDAV_CONFIG', help='Configuration file', metavar='<file>')
@click.option('--name', envvar='CALDAV_NAME', help='Server nickname', metavar='<nickname>')
@click.option('--username', envvar='CALDAV_USERNAME', help='Username on the CalDAV server', metavar='<username>')
@click.option('--passwd', envvar='CALDAV_PASSWD', help='Password on the CalDAV server', metavar='<passwd>')
@click.option('--url', envvar='CALDAV_URL', help='Calendar CalDAV url', metavar='<url>')
@click.option('--timezone', envvar='CALDAV_TIMEZONE', help='Your time zone', metavar='<timezone>')
@click.option('--server', help='Default server (use nickname)', metavar='<nickname>')
@click.option('--calendar', help='Default calendar id', metavar='<uid>')
@click.option('--version', is_flag=True, is_eager=True, expose_value=False, callback=print_version,
              help='caldavctl version')
@click.pass_context
def cli_group(context,
              config_file,
              name, username, passwd, url, timezone,
              server, calendar):
    '''
    caldavctl - command line CalDAV client
    '''

    config = Config(
        config_file,
        name,
        username,
        passwd,
        url,
        timezone,
        server,
        calendar
    )

    context.obj = {
        'config': config,
        'option': []
    }


# CALENDAR

@cli_group.group('calendar', options_metavar='[options]')
@click.pass_context
def calendar_commands(context):
    '''Commands that deal with the calendars on the server'''
    pass


calendar_commands.add_command(calendar.list_calendars)
calendar_commands.add_command(calendar.create_calendar)
calendar_commands.add_command(calendar.delete_calendar)


# EVENTS

@cli_group.group('event', options_metavar='[options]')
@click.pass_context
def event_commands(context):
    '''Event management'''
    pass


event_commands.add_command(event.list_events)
event_commands.add_command(event.create_event)
event_commands.add_command(event.delete_event)


# TODOS

@cli_group.group('todo', options_metavar='[options]')
@click.pass_context
def todo_commands(context):
    '''Todo management'''
    pass


todo_commands.add_command(todo.list_todos)
todo_commands.add_command(todo.create_todo)
todo_commands.add_command(todo.delete_todo)
todo_commands.add_command(todo.toggle_todo_complete)
todo_commands.add_command(todo.percentage_complete)


# JOURNALS

@cli_group.group('journal', options_metavar='[options]')
@click.pass_context
def journal_commands(context):
    '''Journal management'''
    pass


journal_commands.add_command(journal.list_journals)


# ICS

@cli_group.group('ics', options_metavar='[options]')
@click.pass_context
def ics_commands(context):
    '''iCalendar file operations'''
    pass


ics_commands.add_command(ics.get_todo)
ics_commands.add_command(ics.create_todo)
ics_commands.add_command(ics.get_event)


# Backup

@cli_group.group('br', options_metavar='[options]')
@click.pass_context
def backup_and_restore_commands(context):
    '''Backup or restore a calendar'''
    pass


backup_and_restore_commands.add_command(backup.backup_calendar)
backup_and_restore_commands.add_command(backup.restore_calendar)


# Utils

@cli_group.group('utils', options_metavar='[options]')
@click.pass_context
def utilities(context):
    '''Utility commands'''
    pass


utilities.add_command(utils.list_timezones)


def main():
    cli_group()



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

import click
import icalendar
from caldav.lib import error

from caldavctl import dav, get_name, get_version

# Commands:


@click.command('backup')
@click.option('-f', '--file', type=click.File('w'), default=sys.stdout, help='Output file')
@click.pass_obj
def backup_calendar(context, file):
    '''
    Backup calendar to iCalendar file (ics)

    This command outputs the calendar's backup to stdout. Optionally the output
    can be redirected to a file using the -f/--file option.
    '''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    calendar_bck = icalendar.Calendar()
    calendar_bck.add('version', '2.0')
    calendar_bck.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')

    with dav.caldav_calendar(server, calendar_id) as calendar:
        for obj in calendar.objects(load_objects=True):
            # Parse the event's data into a Calendar object
            event_ical = icalendar.Calendar.from_ical(obj.icalendar_instance.to_ical())
            for component in event_ical.walk():
                if component.name in ('VEVENT', 'VTODO', 'VJOURNAL'):  # Filter desired components
                    calendar_bck.add_component(component)

    file.write(str(calendar_bck.to_ical(), 'utf-8'))


@click.command('restore')
@click.option('-f', '--file', type=click.File('r'), default=sys.stdin, help='Input file')
@click.pass_obj
def restore_calendar(context, file):
    '''
    Restore calendar from an iCalendar file (ics)

    This command restores an ics backup file to the selected calendar. It reads
    its input form stdin. Optionally it can restore the backup from a file
    using the -f/--file option.
    '''
    # Read the ICS file
    ics_content = file.read()

    # Parse the ICS file
    calendar_rst = icalendar.Calendar.from_ical(ics_content)

    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        for component in calendar_rst.walk():
            if component.name in ('VEVENT', 'VTODO', 'VJOURNAL'):
                obj = icalendar.Calendar()
                obj.add('version', '2.0')
                obj.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')
                obj.add_component(component)
                obj_string = obj.to_ical().decode('utf-8')
                try:
                    if component.name == 'VEVENT':
                        calendar.save_event(obj_string)
                    elif component.name == 'VTODO':
                        calendar.save_todo(obj_string)
                    elif component.name == 'VJOURNAL':
                        calendar.save_journal(obj_string)
                except error.AuthorizationError:
                    print(f'Error restoring object with uid {component["uid"]}. Skipping.')

    print('Restored.')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
Utilities
'''

import zoneinfo

import click


@click.command('list-timezones', options_metavar='[options]')
@click.option('-s', '--search',
              default=None, metavar='<query>', type=str,
              help='Search time zone')
@click.pass_obj
def list_timezones(context, search):
    '''
    List available time zones

    This command display the available time zones in python's zoneinfo module.
    Use the -s/--search option to search for a specific time zone.
    '''
    timezones = zoneinfo.available_timezones()

    if search:
        result = []
        for tz in timezones:
            if search.lower() in tz.lower():
                result.append(tz)
        if not result:
            result.append(f'No time zone "{search}" found.')
    else:
        result = list(timezones)

    result.sort()

    for tz in result:
        click.echo(tz)
